"""Utilities module."""

from __future__ import annotations

import asyncio
from collections.abc import Callable
from datetime import datetime, time, timedelta, timezone
from typing import Any

from .exceptions import SpaMessageError

MESSAGE_DELIMETER_BYTE = b"~"
MESSAGE_DELIMETER = MESSAGE_DELIMETER_BYTE[0]


def byte_parser(
    value: int,
    offset: int = 0,
    count: int = 8,
    bits: int = 1,
    fn: Callable[[int], int] = lambda _: _,  # pylint: disable=invalid-name
) -> list[int]:
    """Parse a byte."""
    return [
        fn(value >> i * bits + offset & int("0b" + "1" * bits, 2)) for i in range(count)
    ]


def calculate_checksum(data: bytes) -> int:
    """Calculate the checksum byte for a message."""
    crc = 0xB5
    for _, cur in enumerate(data):
        for i in range(8):
            bit = crc & 0x80
            crc = ((crc << 1) & 0xFF) | ((cur >> (7 - i)) & 0x01)
            if bit:
                crc = crc ^ 0x07
        crc &= 0xFF
    for i in range(8):
        bit = crc & 0x80
        crc = (crc << 1) & 0xFF
        if bit:
            crc ^= 0x07
    return crc ^ 0x02


def calculate_time(base_time: time | None, duration: timedelta | None) -> time | None:
    """Calculate the time after adding a duration to a base time."""
    if base_time is None:
        return None
    duration = duration or timedelta()
    return (datetime.combine(datetime.now(), base_time) + duration).time()


def calculate_time_difference(start: time, end: time) -> int:
    """Calculate the difference (in minutes) between a start and end time."""
    return ((end.hour - start.hour) * 60 + (end.minute - start.minute)) % (24 * 60)


async def cancel_task(task: asyncio.Task | None) -> None:
    """Cancel a task."""
    if task is not None and not task.done():
        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            pass


def default(value: Any, default_value: Any | Callable[[], Any]) -> Any:
    """Return value if not None, else default."""
    if value is not None:
        return value
    return default_value() if callable(default_value) else default_value


async def read_one_message(reader: asyncio.StreamReader, timeout: int = 15) -> bytes:
    """Read one message."""
    data = await asyncio.wait_for(reader.readexactly(2), timeout)
    if data[0] != MESSAGE_DELIMETER or data[1] == 0:
        # something went wrong reading a message, so
        # read to the next delimeter and discard
        data += await asyncio.wait_for(
            reader.readuntil(MESSAGE_DELIMETER_BYTE), timeout
        )
        raise SpaMessageError(f"Invalid message: {data.hex()}")
    data = data[1:] + (await reader.readexactly(data[1]))[:-1]
    if data[0] != len(data):
        raise SpaMessageError(f"Incomplete message: {data.hex()}")
    if calculate_checksum(data[:-1]) != data[-1]:
        raise SpaMessageError(f"Invalid checksum: {data.hex()}")
    return data


def to_celsius(fahrenheit: float) -> float:
    """Convert a Fahrenheit temperature to Celsius."""
    return 0.5 * round(((fahrenheit - 32) / 1.8) / 0.5)


def utcnow() -> datetime:
    """Get now in UTC time."""
    return datetime.now(timezone.utc)



"""Balboa spa client."""

from __future__ import annotations

import asyncio
import logging
from datetime import datetime, time, timedelta
from random import uniform
from typing import Any, Callable, TypeVar, cast

from .control import EVENT_UPDATE, EventMixin, FaultLog, HeatModeSpaControl, SpaControl
from .discovery import async_discover
from .enums import (
    AccessibilityType,
    ControlType,
    HeatState,
    LowHighRange,
    MessageType,
    SettingsCode,
    SpaState,
    TemperatureUnit,
    ToggleItemCode,
    WiFiState,
)
from .exceptions import (
    SpaConfigurationNotLoadedError,
    SpaConnectionError,
    SpaMessageError,
)
from .utils import (
    byte_parser,
    calculate_checksum,
    calculate_time,
    calculate_time_difference,
    cancel_task,
    default,
    read_one_message,
    to_celsius,
    utcnow,
)

_LOGGER = logging.getLogger(__name__)
_T = TypeVar("_T")

DEFAULT_PORT = 4257
MESSAGE_DELIMETER_BYTE = b"~"
MESSAGE_DELIMETER = MESSAGE_DELIMETER_BYTE[0]
MESSAGE_SEND = [0x0A, 0xBF]

ACCESSIBILITY_TYPE_MAP = {
    16: AccessibilityType.PUMP_LIGHT,
    32: AccessibilityType.NONE,
    48: AccessibilityType.NONE,
}


class SpaClient(EventMixin):
    """Spa client."""

    def __init__(
        self, host: str, port: int = DEFAULT_PORT, *, mac_address: str | None = None
    ) -> None:
        """Initialize a spa client."""
        self._host = host
        self._port = port

        self._device_configuration_loaded = False
        self._filter_cycle_loaded = False
        self._module_identification_loaded = False
        self._setup_parameters_loaded = False
        self._system_information_loaded = False
        self._configuration_loaded: asyncio.Event = asyncio.Event()

        self._last_log_mesage: bytes | None = None
        self._previous_status: bytes | None = None
        self._last_message_received: datetime | None = None
        self._last_message_sent: datetime | None = None

        self._disconnect = False
        self._reader: asyncio.StreamReader | None = None
        self._writer: asyncio.StreamWriter | None = None
        self._connection_monitor: asyncio.Task | None = None
        self._listener: asyncio.Task | None = None

        self._controls: list[SpaControl] = [
            HeatModeSpaControl(self),
            SpaControl(self, ControlType.TEMPERATURE_RANGE, list(LowHighRange)),
        ]

        # module identification
        self._idigi_device_id: str | None = None
        self._mac_address: str | None = mac_address

        # system information
        self._dip_switch: str | None = None
        self._configuration_signature: str | None = None
        self._current_setup: int | None = None
        self._heater_type: str | None = None
        self._model: str | None = None
        self._software_version: str | None = None
        self._voltage: int | None | None = None

        # setup parameters
        self._low_range: tuple[tuple[int, int], tuple[float, float]] = (
            (50, 99),
            (10.0, 37.0),
        )
        self._high_range: tuple[tuple[int, int], tuple[float, float]] = (
            (80, 104),
            (26.5, 40.0),
        )
        self._pump_count: int = 0

        # filter cycle
        self._filter_cycle_1_start: time | None = None
        self._filter_cycle_1_duration: timedelta = timedelta()
        self._filter_cycle_1_end: time | None = None
        self._filter_cycle_2_enabled: bool = False
        self._filter_cycle_2_start: time | None = None
        self._filter_cycle_2_duration: timedelta = timedelta()
        self._filter_cycle_2_end: time | None = None

        # status update
        self._accessibility_type: AccessibilityType = AccessibilityType.NONE
        self._filter_cycle_1_running: bool = False
        self._filter_cycle_2_running: bool = False
        self._heat_state: HeatState = HeatState.OFF
        self._is_24_hour: bool = False
        self._state: SpaState = SpaState.UNKNOWN
        self._time_hour: int = 0
        self._time_minute: int = 0
        self._time_offset: timedelta = timedelta(0)
        self._temperature_unit: TemperatureUnit = TemperatureUnit.FAHRENHEIT
        self._temperature: float | None = None
        self._target_temperature: float | None = None
        self._temperature_range: int = 0
        self._wifi_state: WiFiState | None = None

        # fault log
        self._fault: FaultLog | None = None

    def _require_configured(self, value: _T | None) -> _T:
        """Ensure the given value is set before returning it, otherwise raise an error."""
        if value is None:
            raise SpaConfigurationNotLoadedError
        return value

    @property
    def host(self) -> str:
        """Return the host address."""
        return self._host

    @property
    def available(self) -> bool:
        """Return True if the client is connected and available."""
        if self.connected and self.last_message_received is not None:
            return self.last_message_received >= utcnow() - timedelta(seconds=15)
        return False

    @property
    def connected(self) -> bool:
        """Return `True` if the client is connected."""
        if self._writer is None:
            return False
        return self._writer.transport.is_reading()  # type: ignore

    @property
    def last_message_received(self) -> datetime | None:
        """Return the last message received datetime."""
        return self._last_message_received

    @property
    def configuration_signature(self) -> str | None:
        """Return the configuration signature."""
        return self._configuration_signature

    @property
    def controls(self) -> list[SpaControl]:
        """Return the controls available."""
        return self._controls

    @property
    def current_setup(self) -> int | None:
        """Return the current setup."""
        return self._current_setup

    @property
    def dip_switch(self) -> str | None:
        """Return the dip switch settings."""
        return self._dip_switch

    @property
    def fault(self) -> FaultLog | None:
        """Return the last received fault."""
        return self._fault

    @property
    def filter_cycle_1_start(self) -> time | None:
        """Return filter cycle 1 start time."""
        return self._filter_cycle_1_start

    @property
    def filter_cycle_1_duration(self) -> timedelta:
        """Return filter cycle 1 duration."""
        return self._filter_cycle_1_duration

    @property
    def filter_cycle_1_end(self) -> time | None:
        """Return filter cycle 1 end time."""
        return self._filter_cycle_1_end

    @property
    def filter_cycle_1_running(self) -> bool:
        """Return `True` if filter cycle 1 is running."""
        return self._filter_cycle_1_running

    @property
    def filter_cycle_2_enabled(self) -> bool:
        """Return `True` if filter cycle 2 is enabled."""
        return self._filter_cycle_2_enabled

    @property
    def filter_cycle_2_start(self) -> time | None:
        """Return filter cycle 2 start time."""
        return self._filter_cycle_2_start

    @property
    def filter_cycle_2_duration(self) -> timedelta:
        """Return filter cycle 2 duration."""
        return self._filter_cycle_2_duration

    @property
    def filter_cycle_2_end(self) -> time | None:
        """Return filter cycle 2 end time."""
        return self._filter_cycle_2_end

    @property
    def filter_cycle_2_running(self) -> bool:
        """Return `True` if filter cycle 2 is running."""
        return self._filter_cycle_2_running

    @property
    def heat_state(self) -> HeatState:
        """Return the heat state."""
        return self._heat_state

    @property
    def heater_type(self) -> str | None:
        """Return the heater type."""
        return self._heater_type

    @property
    def idigi_device_id(self) -> str | None:
        """Return the iDigi Device Id."""
        return self._idigi_device_id

    @property
    def mac_address(self) -> str:
        """Return the MAC address."""
        return self._require_configured(self._mac_address)

    @property
    def model(self) -> str:
        """Return the model."""
        return self._require_configured(self._model)

    @property
    def pump_count(self) -> int:
        """Return the number of pumps."""
        return self._pump_count

    @property
    def software_version(self) -> str | None:
        """Return the software version."""
        return self._software_version

    @property
    def state(self) -> SpaState:
        """Return the spa state."""
        return self._state

    @property
    def temperature_unit(self) -> TemperatureUnit:
        """Return the temperatre unit."""
        return self._temperature_unit

    @property
    def temperature(self) -> float | None:
        """Return the temperature."""
        return self._temperature

    @property
    def target_temperature(self) -> float:
        """Return the target temperature."""
        return self._require_configured(self._target_temperature)

    @property
    def temperature_minimum(self) -> float:
        """Return the temperature minimum."""
        valid_temps = (self._low_range, self._high_range)[self._temperature_range]
        return valid_temps[self._temperature_unit][0]

    @property
    def temperature_maximum(self) -> float:
        """Return the temperature maximum."""
        valid_temps = (self._low_range, self._high_range)[self._temperature_range]
        return valid_temps[self._temperature_unit][1]

    @property
    def time_hour(self) -> int:
        """Return the hour."""
        return self._time_hour

    @property
    def time_minute(self) -> int:
        """Return the minute."""
        return self._time_minute

    @property
    def is_24_hour(self) -> bool:
        """Return `True` if 24-hour time.."""
        return self._is_24_hour

    @property
    def voltage(self) -> int | None:
        """Return the voltage."""
        return self._voltage

    @property
    def aux(self) -> list[SpaControl]:
        """Return the aux controls."""
        return self.get_controls(ControlType.AUX)

    @property
    def blowers(self) -> list[SpaControl]:
        """Return the blower controls."""
        return self.get_controls(ControlType.BLOWER)

    @property
    def circulation_pump(self) -> SpaControl | None:
        """Return the circulation pump control."""
        return next(iter(self.get_controls(ControlType.CIRCULATION_PUMP)), None)

    @property
    def heat_mode(self) -> SpaControl:
        """Return the heat mode control."""
        return self.get_controls(ControlType.HEAT_MODE)[0]

    @property
    def lights(self) -> list[SpaControl]:
        """Return the light controls."""
        return self.get_controls(ControlType.LIGHT)

    @property
    def misters(self) -> list[SpaControl]:
        """Return the mister controls."""
        return self.get_controls(ControlType.MISTER)

    @property
    def pumps(self) -> list[SpaControl]:
        """Return the pump controls."""
        return self.get_controls(ControlType.PUMP)

    @property
    def temperature_range(self) -> SpaControl:
        """Return the temperature range controls."""
        return self.get_controls(ControlType.TEMPERATURE_RANGE)[0]

    def get_controls(self, control_type: ControlType) -> list[SpaControl]:
        """Get controls based on control type."""
        return [
            control for control in self.controls if control.control_type == control_type
        ]

    @property
    def configuration_loaded(self) -> bool:
        """Return `True` if the configuration is loaded."""
        return self._configuration_loaded.is_set()

    def get_current_time(self) -> datetime:
        """Return the current time."""
        return datetime.now() + self._time_offset

    async def async_configuration_loaded(self, timeout: float = 15) -> bool:
        """Wait for configuration to complete."""
        if self.configuration_loaded:
            return True
        try:
            return await asyncio.wait_for(self._configuration_loaded.wait(), timeout)
        except asyncio.TimeoutError:
            return False

    def _check_configuration_loaded(self) -> None:
        """Return `True` if the spa is fully configured."""
        if all(
            (
                self._device_configuration_loaded,
                self._filter_cycle_loaded,
                self._module_identification_loaded,
                self._setup_parameters_loaded,
                self._system_information_loaded,
                self._previous_status,
            )
        ):
            assert self._previous_status
            self._parse_status_update(self._previous_status, True)
            self._configuration_loaded.set()

    async def connect(self) -> bool:
        """Connect to the spa."""
        self._disconnect = False
        return await self._connect()

    async def _connect(self) -> bool:
        """Connect to the spa."""
        if self.connected:
            _LOGGER.debug("%s -- already connected", self._host)
            return True
        if self._disconnect:
            _LOGGER.debug(
                "%s -- connect skipped due to previous disconnect request", self._host
            )
            return False

        _LOGGER.debug("%s -- establishing connection", self._host)
        try:
            self._reader, self._writer = await asyncio.wait_for(
                asyncio.open_connection(self._host, self._port), 10
            )
        except (
            asyncio.TimeoutError,
            ConnectionRefusedError,
            TimeoutError,
            OSError,
        ) as err:
            msg = "Timed out" if isinstance(err, asyncio.TimeoutError) else err
            _LOGGER.error("%s ## cannot connect: %s", self._host, msg)
        except Exception as ex:  # pylint: disable=broad-except
            _LOGGER.error("%s ## error connecting: %s", self._host, ex)
        else:
            _LOGGER.debug("%s -- connected", self._host)
            self._listener = asyncio.ensure_future(self._start_listener())
            asyncio.ensure_future(self.request_all_configuration(True))
            await cancel_task(self._connection_monitor)

            async def _monitor() -> None:
                attempt = 0
                while not self._disconnect:
                    while self.connected:
                        await asyncio.sleep(1)
                    if not await self._connect():
                        await asyncio.sleep(min(1 * 2**attempt + uniform(0, 1), 60))
                        attempt += 1

            self._connection_monitor = asyncio.ensure_future(_monitor())
        return self.connected

    async def disconnect(self) -> None:
        """Disconnect from the spa."""
        _LOGGER.debug("%s -- disconnect requested", self._host)
        self._disconnect = True
        await cancel_task(self._connection_monitor)
        if self._writer is not None:
            self._writer.close()
            try:
                await self._writer.wait_closed()
            except Exception:  # pylint: disable=broad-except
                pass
        await cancel_task(self._listener)
        self._reader = self._writer = None
        _LOGGER.debug("%s -- disconnected", self._host)

    async def _start_listener(self) -> None:
        """Start the listener."""
        timeout = 15
        wait_time = timedelta(seconds=timeout)
        assert self._reader
        while self.connected:
            try:
                data = await read_one_message(self._reader, timeout)
            except SpaMessageError as err:
                _LOGGER.debug("%s ## %s", self._host, err)
                continue
            except (asyncio.TimeoutError, asyncio.IncompleteReadError):
                if not (sent := self._last_message_sent) or sent + wait_time < utcnow():
                    self.emit(EVENT_UPDATE)
                    await self.send_device_present()
                continue
            except Exception as ex:  # pylint: disable=broad-except
                _LOGGER.error("%s ## %s", self._host, ex)
                continue
            self._process_message(data)
        self.emit(EVENT_UPDATE)
        _LOGGER.debug("%s -- stopped listening", self._host)

    def _process_message(self, data: bytes) -> None:
        """Process a message."""
        self._last_message_received = utcnow()
        message_type = self._log_message(data)
        data = data[4:-1]

        if message_type == MessageType.STATUS_UPDATE:
            self._parse_status_update(data)
        elif message_type == MessageType.MODULE_IDENTIFICATION:
            self._parse_module_identification(data)
        elif message_type == MessageType.FILTER_CYCLE:
            self._parse_filter_cycle(data)
        elif message_type == MessageType.FAULT_LOG:
            self._parse_fault_log(data)
        elif message_type == MessageType.DEVICE_CONFIGURATION:
            self._parse_device_configuration(data)
        elif message_type == MessageType.SETUP_PARAMETERS:
            self._parse_setup_parameters(data)
        elif message_type == MessageType.SYSTEM_INFORMATION:
            self._parse_system_information(data)

    def _parse_device_configuration(self, data: bytes) -> None:
        """Parse a device configuration message.

        Device configuration messages have a length of 6 bytes with the following information:

        Byte  | Data
        ---------------------------
        00    | P4P3P2P1 - Pumps 1-4
        01    | P6P7P8P5 - Pumps 5-8
        02    | L4L3L2L1 - Lights 1-4
        03    | CxxxB2B1 - circulation pump, blowers 1-2
        04    | xMMMAAAA - mister 1-3, aux 1-4
        05    | ?
        """
        if not self._device_configuration_loaded:

            def _add_controls(control_type: ControlType, on_states: list[int]) -> None:
                self._controls.extend(
                    SpaControl(
                        self,
                        control_type,
                        state + 1,
                        index if len(on_states) > 1 else None,
                    )
                    for index, state in enumerate(on_states)
                    if state > 0
                )

            pumps = [
                *byte_parser(data[0], count=4, bits=2),  # pumps 1-4
                data[1] & 0x03,  # pump 5
                data[1] >> 6 & 0x03,  # pump 6
                data[1] >> 4 & 0x03,  # pump 7
                data[1] >> 2 & 0x03,  # pump 8
            ]
            lights = byte_parser(data[2], count=4, bits=2)
            circulation_pump = data[3] >> 7  # only one
            blowers = byte_parser(data[3], count=2, bits=2)
            auxs = byte_parser(data[4], count=4)
            misters = byte_parser(data[4], offset=4, count=3)

            _add_controls(ControlType.PUMP, pumps)
            _add_controls(ControlType.LIGHT, lights)
            _add_controls(ControlType.CIRCULATION_PUMP, [circulation_pump])
            _add_controls(ControlType.BLOWER, blowers)
            _add_controls(ControlType.AUX, auxs)
            _add_controls(ControlType.MISTER, misters)

            self._device_configuration_loaded = True
            self._check_configuration_loaded()

    def _parse_fault_log(self, data: bytes) -> None:
        """Parse a fault log message.

        Fault log messages have a length of 10 bytes with the following information:

        Byte  | Data
        ---------------------------
        00    | fault count
        01    | entry number
        02    | message code
        03    | days ago
        04    | time hour
        05    | time minute
        06    | flags
        07    | target temperature
        08    | sensor A temperature
        09    | sensor B temperature
        """
        self._fault = FaultLog(*(*data, self.get_current_time()))

    def _parse_filter_cycle(self, data: bytes) -> None:
        """Parse a filter cycle message.

        Filter cycle messages have a length of 8 bytes with the following information:

        Byte  | Data
        ---------------------------
        00    | filter cycle 1 start hour
        01    | filter cycle 1 start minute
        02    | filter cycle 1 duration hours
        03    | filter cycle 1 duration minutes
        04    | filter cycle 2 enabled and start hour
        05    | filter cycle 2 start minute
        06    | filter cycle 2 duration hours
        07    | filter cycle 2 duration minutes
        """
        self._filter_cycle_1_start = time(data[0], data[1])
        self._filter_cycle_1_duration = timedelta(hours=data[2], minutes=data[3])
        self._filter_cycle_1_end = calculate_time(
            self._filter_cycle_1_start, self._filter_cycle_1_duration
        )

        self._filter_cycle_2_enabled = bool(data[4] >> 7)
        self._filter_cycle_2_start = time(data[4] & 0x7F, data[5])
        self._filter_cycle_2_duration = timedelta(hours=data[6], minutes=data[7])
        self._filter_cycle_2_end = calculate_time(
            self._filter_cycle_2_start, self._filter_cycle_2_duration
        )

        self._filter_cycle_loaded = True
        self._check_configuration_loaded()

    def _parse_module_identification(self, data: bytes) -> None:
        """Parse a module identification message.

        Module identification messages have a length of 25 bytes with the following information:

        Byte  | Data
        ---------------------------
        00-02 | ? ? ?
        03-08 | mac address
        09-24 | iDigi device id (used to communicate with Balboa cloud API)
        """
        self._mac_address = ":".join(f"{x:02x}" for x in data[3:9])
        idigi_device_id = "-".join(data[i : i + 4].hex() for i in range(9, 25, 4))
        self._idigi_device_id = idigi_device_id.upper()
        self._module_identification_loaded = True
        self._check_configuration_loaded()

    def _parse_setup_parameters(self, data: bytes) -> None:
        """Parse a setup parameters message.

        Setup parameters messages have a length of 9 bytes with the following information:

        Byte  | Data
        ---------------------------
        00-01 | ? ?
        02    | low range minimum temperature in °F
        03    | low range maximum temperature in °F
        04    | high range minimum temperature in °F
        05    | high range maximum temperature in °F
        06    | ?
        07    | pump counter (add the number of "1"s from bit)
        08    | ?
        """
        if not self._setup_parameters_loaded:
            low, high = data[2], data[3]
            self._low_range = ((low, high), (to_celsius(low), to_celsius(high)))
            low, high = data[4], data[5]
            self._high_range = ((low, high), (to_celsius(low), to_celsius(high)))
            self._pump_count = sum(byte_parser(data[7]))
            self._setup_parameters_loaded = True
            self._check_configuration_loaded()

    def _parse_status_update(self, data: bytes, reprocess: bool = False) -> None:
        """Parse a status update message.

        Status update messages have a length of 24 bytes with the following information:

        Byte  | Data
        ---------------------------
        00    | spa state ?
        01    | initialization mode ?
        02    | current temperature
        03    | current hour
        04    | current minute
        05    | heat mode
        06-08 | ? ? ?
        09    | temperature scale, time format, filter mode, accessibility type
        10    | temperature range, heating
        11    | pumps 1-4
        12    | pumps 5-8
        13    | circulation pump, blower state
        14    | lights 1-4
        15    | mister 1-3, aux 1-4
        16-19 | ? ? ? ?
        20    | target temperature
        21    | ?
        22    | wifi
        23    | ?
        """
        if data == self._previous_status and not reprocess:
            # No new information, so ignore it
            return

        self._previous_status = data
        self._state = SpaState(data[0])
        self._time_hour = data[3]
        self._time_minute = data[4]
        if not reprocess:
            now = datetime.now()
            device_time = now.replace(hour=self._time_hour, minute=self._time_minute)
            self._time_offset = device_time - now
        self._is_24_hour = (flag := data[9]) & 0x02 != 0
        if flag & 0x01 == 0:
            self._temperature_unit = TemperatureUnit.FAHRENHEIT
            divisor = 1
        else:
            self._temperature_unit = TemperatureUnit.CELSIUS
            divisor = 2
        temperature = None if (temperature := data[2]) == 255 else temperature / divisor
        self._temperature = temperature
        self._target_temperature = data[20] / divisor
        self._filter_cycle_1_running = flag & 0x04 != 0
        self._filter_cycle_2_running = flag & 0x08 != 0
        self._accessibility_type = ACCESSIBILITY_TYPE_MAP.get(
            flag & 0x48, AccessibilityType.ALL
        )
        self._temperature_range = ((flag := data[10]) >> 2) & 0x01
        self._update_control_states(
            ControlType.TEMPERATURE_RANGE, [self._temperature_range]
        )
        self._heat_state = HeatState(flag >> 4 & 0x03)
        light_states = byte_parser(data[14], count=4, bits=2, fn=lambda _: _ >> 1)
        self._update_control_states(ControlType.LIGHT, light_states)
        heat_mode = data[5] & 0x03
        self._update_control_states(ControlType.HEAT_MODE, [heat_mode])
        pump_states = byte_parser(data[11], count=4, bits=2)
        pump_states.extend(byte_parser(data[12], count=4, bits=2))
        self._update_control_states(ControlType.PUMP, pump_states)
        circulation_pump = (data[13] & 0x03) >> 1
        self._update_control_states(ControlType.CIRCULATION_PUMP, [circulation_pump])
        blower_states = byte_parser(data[13], 1, 2, 2)
        self._update_control_states(ControlType.BLOWER, blower_states)
        mister_states = byte_parser(data[15], count=3)
        self._update_control_states(ControlType.MISTER, mister_states)
        aux_states = byte_parser(data[15], offset=3, count=4)
        self._update_control_states(ControlType.AUX, aux_states)
        self._wifi_state = WiFiState(int((data[22] & 0xF0) / 16))

        if not self.configuration_loaded and not reprocess:
            self._check_configuration_loaded()

        self.emit(EVENT_UPDATE)

    def _update_control_states(
        self, control_type: ControlType, states: list[int]
    ) -> None:
        """Update the control states."""
        for index, state in enumerate(states):
            if control := next(
                (
                    control
                    for control in self.controls
                    if control.control_type == control_type
                    and (control.index == index or control.index is None)
                ),
                None,
            ):
                control.update(state)

    def _parse_system_information(self, data: bytes) -> None:
        """Parse a system information message.

        System information messages have a length of 21 bytes with the following information:

        Byte  | Data
        ---------------------------
        00-03 | software id (ssid) and version
        04-11 | model name
        12    | current setup
        13-16 | configuration signature
        17    | voltage
        18    | heater type
        19-20 | dip switch
        """
        self._software_version = f"M{data[0]}_{data[1]} V{data[2]}.{data[3]}"
        self._model = "".join(map(chr, data[4:12])).strip()
        self._current_setup = data[12]
        self._configuration_signature = data[13:17].hex()
        self._voltage = 240 if data[17] == 0x01 else None
        self._heater_type = "standard" if data[18] == 0x0A else "unknown"
        self._dip_switch = f"{data[19]:08b}{data[20]:08b}"
        self._system_information_loaded = True
        self._check_configuration_loaded()

    def _log_message(self, data: bytes) -> MessageType:
        """Log message and return message type."""
        message_type = MessageType(data[3])
        if self._last_log_mesage != data:
            self._last_log_mesage = data
            _LOGGER.debug("%s -> %s: %s", self._host, message_type.name, data.hex())
        return message_type

    async def request_all_configuration(self, wait: bool = False) -> None:
        """Request the full spa configuration."""
        if not self._module_identification_loaded or not wait:
            await self.request_module_identification()
        if not self._system_information_loaded or not wait:
            await self.request_system_information()
        if not self._setup_parameters_loaded or not wait:
            await self.request_setup_parameters()
        if not self._device_configuration_loaded or not wait:
            await self.request_device_configuration()
        if not self._filter_cycle_loaded or not wait:
            await self.request_filter_cycle()
        if wait and not await self.async_configuration_loaded(3):
            if self.connected:
                await self.request_all_configuration(wait)

    async def request_device_configuration(self) -> None:
        """Request the device configuration."""
        await self.send_message(
            MessageType.REQUEST, SettingsCode.DEVICE_CONFIGURATION, 0x00, 0x01
        )

    async def request_fault_log(self, entry: int = 0xFF) -> None:
        """Request a fault log entry.

        entry: The fault log to retrieve, 0..23 or 0xFF (255) for the last fault
        """
        if not 0 <= entry < 24 and entry != 0xFF:
            raise ValueError(
                f"Invalid fault log entry: {entry} (expected 0–23 or 0xFF for the last fault)"
            )

        await self.send_message(
            MessageType.REQUEST, SettingsCode.FAULT_LOG, entry % 256, 0x00
        )

    async def request_filter_cycle(self) -> None:
        """Request the filter cycle."""
        await self.send_message(
            MessageType.REQUEST, SettingsCode.FILTER_CYCLE, 0x00, 0x00
        )

    async def request_module_identification(self) -> None:
        """Request the module identification."""
        await self.send_device_present()

    async def request_setup_parameters(self) -> None:
        """Request the system information."""
        await self.send_message(
            MessageType.REQUEST, SettingsCode.SETUP_PARAMETERS, 0x00, 0x00
        )

    async def request_system_information(self) -> None:
        """Request the system information."""
        await self.send_message(
            MessageType.REQUEST, SettingsCode.SYSTEM_INFORMATION, 0x00, 0x00
        )

    async def send_device_present(self) -> None:
        """Send a device present message."""
        await self.send_message(MessageType.DEVICE_PRESENT)

    async def send_message(
        self, message_type: MessageType | None, *message: int
    ) -> None:
        """Send a message to the spa with variable length."""
        if not self.connected:
            return
        if not message_type:
            message_type = MessageType.UNKNOWN
        prefix = [*MESSAGE_SEND, message_type.value] if message_type else []
        message_data = [*prefix, *message]
        message_length = len(message_data) + 2
        data = bytearray(message_length + 2)
        data[0] = MESSAGE_DELIMETER
        data[1] = message_length
        data[2:message_length] = message_data
        data[-2] = calculate_checksum(data[1:message_length])
        data[-1] = MESSAGE_DELIMETER

        _LOGGER.debug(
            "%s <- %s%s: %s",
            self._host,
            message_type.name,
            f"_{SettingsCode(data[5]).name}"
            if message_type == MessageType.REQUEST
            else "",
            data[1:-1].hex(),
        )
        try:
            assert self._writer
            self._writer.write(data)
            await self._writer.drain()
            self._last_message_sent = utcnow()
        except Exception as ex:  # pylint: disable=broad-except
            _LOGGER.error("%s ## error sending message: %s", self._host, ex)

    async def __aenter__(self) -> SpaClient:
        """Connect and start listening for messages."""
        if not await self._connect():
            raise SpaConnectionError()
        return self

    async def __aexit__(self, *exctype: Any) -> None:
        """Disconnect."""
        await self.disconnect()

    async def configure_filter_cycle(
        self,
        filter_cycle: int,
        *,
        start: time | None = None,
        end: time | None = None,
        duration: timedelta | None = None,
        enabled: bool | None = None,
    ) -> None:
        """Configure a filter cycle."""
        if filter_cycle not in (1, 2):
            raise ValueError(f"Invalid filter cycle: {filter_cycle} (expected 1 or 2)")
        if all(value is None for value in (start, end, duration, enabled)):
            raise ValueError(
                "At least one of start, end, duration, or enabled must be provided"
            )
        if duration is not None and end is not None:
            raise ValueError("Only one of end or duration should be provided")
        if filter_cycle == 1:
            if all(value is None for value in (start, end, duration)):
                raise ValueError(
                    "Filter cycle 1 requires at least one of start, end, or duration"
                )
            if enabled is False:
                raise ValueError("Filter cycle 1 cannot be disabled")
        if duration is not None and not timedelta(minutes=15) <= duration <= timedelta(
            hours=24
        ):
            raise ValueError(
                f"Invalid duration: {duration} (must be between 15 minutes and 24 hours)"
            )

        start = start or getattr(self, f"filter_cycle_{filter_cycle}_start")
        if start is None:
            raise ValueError("A valid start time could not be determined")

        if end is None:
            if duration is None:
                end = getattr(self, f"filter_cycle_{filter_cycle}_end")
            else:
                end = calculate_time(start, duration)

        if end is None:
            raise ValueError("A valid end time/duration could not be determined")

        # 1440 = 24 hours
        minutes = 1440 if start == end else calculate_time_difference(start, end)

        if not 15 <= minutes <= 1440:  # one more sanity check
            raise ValueError(
                f"Invalid duration: {timedelta(minutes=minutes)} (must be between 15 minutes and 24 hours)"
            )

        duration_hours, duration_minutes = divmod(minutes, 60)

        params: dict[str, Any] = {
            f"filter_cycle_{filter_cycle}_hour": start.hour,
            f"filter_cycle_{filter_cycle}_minute": start.minute,
            f"filter_cycle_{filter_cycle}_duration_hours": duration_hours,
            f"filter_cycle_{filter_cycle}_duration_minutes": duration_minutes,
        }

        if filter_cycle == 2 and enabled is not None:
            params["filter_cycle_2_enabled"] = enabled

        await self.set_filter_cycle(**params)

    async def set_filter_cycle(
        self,
        filter_cycle_1_hour: int | None = None,
        filter_cycle_1_minute: int | None = None,
        filter_cycle_1_duration_hours: int | None = None,
        filter_cycle_1_duration_minutes: int | None = None,
        filter_cycle_2_enabled: bool | None = None,
        filter_cycle_2_hour: int | None = None,
        filter_cycle_2_minute: int | None = None,
        filter_cycle_2_duration_hours: int | None = None,
        filter_cycle_2_duration_minutes: int | None = None,
    ) -> None:
        """Set the filter cycle."""
        values = (
            filter_cycle_1_hour,
            filter_cycle_1_minute,
            filter_cycle_1_duration_hours,
            filter_cycle_1_duration_minutes,
            filter_cycle_2_enabled,
            filter_cycle_2_hour,
            filter_cycle_2_minute,
            filter_cycle_2_duration_hours,
            filter_cycle_2_duration_minutes,
        )
        if all(value is None for value in values):
            return

        old_cycle_1_duration = divmod(self.filter_cycle_1_duration.seconds // 60, 60)
        old_cycle_2_duration = divmod(self.filter_cycle_2_duration.seconds // 60, 60)
        enabled = default(filter_cycle_2_enabled, self.filter_cycle_2_enabled) << 7

        def _time_attr(_prop: str, _attr: str) -> int | Callable[[], int]:
            def getter() -> int:
                if (_time := getattr(self, _prop, None)) is None:
                    raise ValueError(
                        f"Unable to get {_attr} from {_prop} because it is {_time}"
                    )
                return cast(int, getattr(_time, _attr))

            return getter

        # fmt: off
        message = [
            default(filter_cycle_1_hour, _time_attr("filter_cycle_1_start", "hour")),
            default(filter_cycle_1_minute, _time_attr("filter_cycle_1_start", "minute")),
            default(filter_cycle_1_duration_hours, old_cycle_1_duration[0]),
            default(filter_cycle_1_duration_minutes, old_cycle_1_duration[1]),
            enabled | default(filter_cycle_2_hour, _time_attr("filter_cycle_2_start", "hour")),
            default(filter_cycle_2_minute, _time_attr("filter_cycle_2_start", "minute")),
            default(filter_cycle_2_duration_hours, old_cycle_2_duration[0]),
            default(filter_cycle_2_duration_minutes, old_cycle_2_duration[1]),
        ]
        # fmt: on

        await self.send_message(MessageType.FILTER_CYCLE, *message)
        await self.request_filter_cycle()

    async def set_temperature(self, temperature: float) -> None:
        """Set the target temperature."""
        valid_temps = (self._low_range, self._high_range)[self._temperature_range]
        low, high = valid_temps[self._temperature_unit]
        if not low <= temperature <= high:
            raise ValueError(
                f"Invalid temperature: {temperature} (expected {low}..{high})"
            )
        if self._temperature_unit == TemperatureUnit.CELSIUS:
            temperature *= 2
        await self.send_message(MessageType.SET_TEMPERATURE, int(temperature))

    async def set_temperature_range(self, temperature_range: LowHighRange) -> None:
        """Set the temperature range."""
        if self._temperature_range == temperature_range:
            return
        await self.send_message(
            MessageType.TOGGLE_STATE, ToggleItemCode.TEMPERATURE_RANGE
        )

    async def set_temperature_unit(self, unit: TemperatureUnit) -> None:
        """Set the temperature unit."""
        await self.send_message(MessageType.SET_TEMPERATURE_UNIT, 0x01, unit.value)

    async def set_time(
        self, hour: int, minute: int, is_24_hour: bool | None = None
    ) -> None:
        """Set the time."""
        try:
            time(hour, minute)
        except ValueError as err:
            raise ValueError(f"Invalid time format: {hour}:{minute}") from err
        if is_24_hour is None:
            is_24_hour = self._is_24_hour
        await self.send_message(MessageType.SET_TIME, (is_24_hour << 7) | hour, minute)

    async def set_24_hour_time(self, is_24_hour: bool) -> None:
        """Set the 24-hour time."""
        await self.set_time(self._time_hour, self._time_minute, is_24_hour)

    @classmethod
    async def discover(
        cls, return_once_found: bool = False, *, timeout: int = 10
    ) -> list[SpaClient]:
        """Discover spas on the network within a specified timeout.

        If return_once_found is True, the first spa found will stop the scan.
        """
        spas = await async_discover(return_once_found, timeout=timeout)
        return [cls(spa.address, mac_address=spa.mac_address) for spa in spas]



"""Enums module."""

from __future__ import annotations

import logging
from enum import Enum, IntEnum

_LOGGER = logging.getLogger(__name__)


class MessageType(IntEnum):
    """Message type."""

    DEVICE_PRESENT = 0x04
    TOGGLE_STATE = 0x11
    STATUS_UPDATE = 0x13
    SET_TEMPERATURE = 0x20
    SET_TIME = 0x21
    REQUEST = 0x22
    FILTER_CYCLE = 0x23
    SYSTEM_INFORMATION = 0x24
    SETUP_PARAMETERS = 0x25
    PREFERENCES = 0x26
    SET_TEMPERATURE_UNIT = 0x27
    FAULT_LOG = 0x28
    DEVICE_CONFIGURATION = 0x2E
    SET_WIFI = 0x92
    MODULE_IDENTIFICATION = 0x94

    UNKNOWN = -1

    @classmethod
    def _missing_(cls, _: object) -> MessageType:
        """Return default if not found."""
        return cls.UNKNOWN


class SettingsCode(IntEnum):
    """Settings code."""

    DEVICE_CONFIGURATION = 0x00
    FILTER_CYCLE = 0x01
    SYSTEM_INFORMATION = 0x02
    SETUP_PARAMETERS = 0x04
    FAULT_LOG = 0x20

    UNKNOWN = -1

    @classmethod
    def _missing_(cls, _: object) -> SettingsCode:
        """Return default if not found."""
        return cls.UNKNOWN


class ControlType(Enum):
    """Control type."""

    AUX = "Aux"
    BLOWER = "Blower"
    CIRCULATION_PUMP = "Circulation pump"
    HEAT_MODE = "Heat mode"
    LIGHT = "Light"
    MISTER = "Mister"
    PUMP = "Pump"
    TEMPERATURE_RANGE = "Temperature range"


class AccessibilityType(IntEnum):
    """Accessibility type."""

    PUMP_LIGHT = 0
    NONE = 1
    ALL = 2


class HeatState(IntEnum):
    """Heat state."""

    OFF = 0
    HEATING = 1
    HEAT_WAITING = 2


class SpaState(IntEnum):
    """Spa state."""

    RUNNING = 0x00
    INITIALIZING = 0x01
    HOLD_MODE = 0x05
    AB_TEMPS_ON = 0x14
    TEST_MODE = 0x17

    UNKNOWN = -1

    @classmethod
    def _missing_(cls, value: object) -> SpaState:
        """Handle unknown values by returning UNKNOWN instead of raising an error."""
        _LOGGER.warning("Received unknown value %s for %s", value, cls.__name__)
        return cls.UNKNOWN


class TemperatureUnit(IntEnum):
    """Tempeature unit."""

    FAHRENHEIT = 0
    CELSIUS = 1


class ToggleItemCode(IntEnum):
    """Toggle item code."""

    NORMAL_OPERATION = 0x01
    CLEAR_NOTIFICATION = 0x03
    PUMP_1 = 0x04
    PUMP_2 = 0x05
    PUMP_3 = 0x06
    PUMP_4 = 0x07
    PUMP_5 = 0x08
    PUMP_6 = 0x09
    BLOWER = 0x0C
    MISTER = 0x0E
    LIGHT_1 = 0x11
    LIGHT_2 = 0x12
    LIGHT_3 = 0x13
    LIGHT_4 = 0x14
    AUX_1 = 0x16
    AUX_2 = 0x17
    SOAK_MODE = 0x1D
    HOLD_MODE = 0x3C
    CIRCULATION_PUMP = 0x3D
    TEMPERATURE_RANGE = 0x50
    HEAT_MODE = 0x51


class WiFiState(IntEnum):
    """Wi-Fi state."""

    OK = 0
    SPA_NOT_COMMUNICATING = 1
    STARTUP = 2
    PRIME = 3
    HOLD = 4
    PANEL = 5


class HeatMode(IntEnum):
    """Heat modes."""

    READY = 0
    REST = 1
    READY_IN_REST = 2


class LowHighRange(IntEnum):
    """Low/high range."""

    LOW = 0
    HIGH = 1


class OffOnState(IntEnum):
    """On/off state."""

    OFF = 0
    ON = 1


class OffLowHighState(IntEnum):
    """Off/low/high state."""

    OFF = 0
    LOW = 1
    HIGH = 2


class OffLowMediumHighState(IntEnum):
    """Off/low/medium/high state."""

    OFF = 0
    LOW = 1
    MEDIUM = 2
    HIGH = 3


class UnknownState(IntEnum):
    """Unknown state."""

    UNKNOWN = -1



"""Main entry."""

import argparse
import asyncio
import logging
import sys
from enum import IntEnum
from typing import Union

try:
    from . import SpaClient, SpaConnectionError, SpaControl, __version__
    from .enums import SpaState
except ImportError:
    from pybalboa import SpaClient, SpaConnectionError, SpaControl, __version__
    from pybalboa.enums import SpaState


async def run_discovery(first_spa: bool = True) -> None:
    """Attempt to discover a spa and try some commands."""
    spas = await SpaClient.discover(first_spa)
    for spa in spas:
        await connect_and_listen(spa=spa)


async def connect_and_listen(
    host: Union[str, None] = None, spa: Union[SpaClient, None] = None
) -> None:
    """Connect to the spa and try some commands."""
    print("******** Testing spa connection and configuration **********")
    try:
        if host:
            spa = SpaClient(host)
        if not spa:
            print("No spa provided")
            return
        async with spa:
            if not await spa.async_configuration_loaded():
                if spa.state == SpaState.TEST_MODE:
                    print("Config not loaded, spa is in test mode!")
                else:
                    print("Config not loaded, something is wrong!")
                return

            print()
            print("Module identification")
            print("---------------------")
            print(f"MAC address: {spa.mac_address}")
            print(f"iDigi Device Id: {spa.idigi_device_id}")
            print()

            print("Device configuration")
            print("--------------------")
            print(spa.circulation_pump)
            print(f"Pumps: {[pump.name for pump in spa.pumps]}")
            print(f"Lights: {[light.name for light in spa.lights]}")
            print(f"Aux: {[aux.name for aux in spa.aux]}")
            print(f"Blower: {[blower.name for blower in spa.blowers]}")
            print(f"Mister: {[mister.name for mister in spa.misters]}")
            print()

            print("System information")
            print("------------------")
            print(f"Model: {spa.model}")
            print(f"Software version: {spa.software_version}")
            print(f"Configuration signature: {spa.configuration_signature}")
            print(f"Current setup: {spa.current_setup}")
            print(f"Voltage: {spa.voltage}")
            print(f"Heater type: {spa.heater_type}")
            print(f"DIP switch: {spa.dip_switch}")
            print()

            print("Setup parameters")
            print("----------------")
            print(f"Min temps: {spa._low_range}")  # pylint: disable=protected-access
            print(f"Max temps: {spa._high_range}")  # pylint: disable=protected-access
            print(f"Pump count: {spa.pump_count}")
            print()

            print("Filter cycle")
            print("------------")
            print(f"Filter cycle 1 start: {spa.filter_cycle_1_start}")
            print(f"Filter cycle 1 duration: {spa.filter_cycle_1_duration}")
            print(
                f"Filter cycle 2 start: {spa.filter_cycle_2_start} ({'en' if spa.filter_cycle_2_enabled else 'dis'}abled)"
            )
            print(f"Filter cycle 2 duration: {spa.filter_cycle_2_duration}")
            print()

            print("Status update")
            print("-------------")
            print(f"Temperature unit: {spa.temperature_unit.name}")
            print(f"Temperature: {spa.temperature}")
            print(f"Target temperature: {spa.target_temperature}")
            print(f"Temperature range: {spa.temperature_range.state.name}")
            print(f"Heat mode: {spa.heat_mode.state.name}")
            print(f"Heat state: {spa.heat_state.name}")
            print(f"Pump status: {spa.pumps}")
            print(spa.circulation_pump)
            print(f"Light status: {spa.lights}")
            print(f"Mister status: {spa.misters}")
            print(f"Aux status: {spa.aux}")
            print(f"Blower status: {spa.blowers}")
            print(
                f"Spa time: {spa.time_hour:02d}:{spa.time_minute:02d} {'24hr' if spa.is_24_hour else '12hr'}"
            )
            print(f"Filter cycle 1 running: {spa.filter_cycle_1_running}")
            print(f"Filter cycle 2 running: {spa.filter_cycle_2_running}")
            print()

            await test_controls(spa)
    except SpaConnectionError:
        print(f"Failed to connect to spa at {host}")
    else:
        print()
        print(
            "If something is not working as expected, please create an issue and add the above output at:"
        )
        print("https://github.com/garbled1/pybalboa/issues/")
        print()


async def test_controls(spa: SpaClient) -> None:
    """Test spa controls."""
    print("******** Testing spa controls **********")
    print()
    print("Temperature control")
    print("-------------------")
    assert spa.target_temperature is not None
    assert spa.temperature_maximum is not None
    assert spa.temperature_minimum is not None
    target_temperature = spa.target_temperature
    await adjust_temperature(
        spa,
        spa.temperature_maximum
        if spa.target_temperature != spa.temperature_maximum
        else spa.temperature_minimum,
    )
    await adjust_temperature(spa, target_temperature)
    print()

    for control in spa.controls:
        print(f"{control.name} control")
        print("-" * (len(control.name) + 8))
        state = control.state
        for option in control.options:
            if option not in (state, control.state):
                await adjust_control(control, option)
        if control.state != state:
            await adjust_control(control, state)
        print()


async def adjust_temperature(spa: SpaClient, temperature: float) -> None:
    """Adjust target temperature settings."""
    print(f"Current target temperature: {spa.target_temperature}")
    print(f"  Set to {temperature}")
    await spa.set_temperature(temperature)

    async def _temperature_check() -> None:
        while spa.target_temperature != temperature:
            await asyncio.sleep(0.1)

    wait = 10
    try:
        await asyncio.wait_for(_temperature_check(), wait)
        print(f"  Set temperature is now {spa.target_temperature}")
    except asyncio.TimeoutError:
        print(
            f"  Set temperature was not changed after {wait} seconds; is {spa.target_temperature}"
        )


async def adjust_control(control: SpaControl, state: IntEnum) -> None:
    """Adjust control state."""
    print(f"Current state: {control.state.name}")
    print(f"  Set to {state.name}")
    if not await control.set_state(state):
        return

    async def _state_check() -> None:
        while control.state != state:
            await asyncio.sleep(0.1)

    wait = 10
    try:
        await asyncio.wait_for(_state_check(), wait)
        print(f"  State is now {control.state.name}")
    except asyncio.TimeoutError:
        print(f"  State was not changed after {wait} seconds; is {control.state.name}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Connect to a spa and listen for updates.",
        usage=f"{sys.argv[0]} [host] [-d | --debug] [--all]",
    )
    parser.add_argument("host", nargs="?", help="Spa IP address or hostname (optional)")
    parser.add_argument(
        "-d", "--debug", action="store_true", help="Enable debug logging"
    )
    parser.add_argument(
        "--all",
        action="store_true",
        help="Discover all available spas instead of just the first one.",
    )

    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    print(f"pybalboa version: {__version__}")

    if args.host:
        asyncio.run(connect_and_listen(args.host))
    else:
        print("No host provided. Running in discovery mode...")
        asyncio.run(run_discovery(first_spa=not args.all))

    sys.exit(0)



"""Balboa spa discovery."""

from __future__ import annotations

import asyncio
import logging
from dataclasses import dataclass
from socket import AF_INET, IPPROTO_UDP
from typing import Any

_LOGGER = logging.getLogger(__name__)

BROADCAST_ADDRESS = ("255.255.255.255", 30303)
BROADCAST_MESSAGE = b"Discovery"
BROADCAST_INTERVAL = 3


async def async_discover(
    return_once_found: bool = False, *, timeout: int = 10
) -> list[DiscoveredSpa]:
    """Discover spas on the network within a specified timeout."""
    loop = asyncio.get_running_loop()
    transport, protocol = await loop.create_datagram_endpoint(
        lambda: SpaDiscoveryProtocol(return_once_found),
        # local_addr=("0.0.0.0", 0),
        family=AF_INET,
        proto=IPPROTO_UDP,
        # reuse_port=True,
        allow_broadcast=True,
    )

    try:
        await asyncio.wait_for(protocol.discovery_complete.wait(), timeout=timeout)
    except asyncio.TimeoutError:
        if not protocol.spas:
            _LOGGER.debug("Discovery timed out")
    finally:
        transport.close()

    return protocol.spas


@dataclass
class DiscoveredSpa:
    """Discovered spa."""

    address: str
    port: int
    mac_address: str
    hostname: str


class SpaDiscoveryProtocol(asyncio.DatagramProtocol):
    """Spa discovery protocol."""

    def __init__(self, return_once_found: bool = False) -> None:
        """Initialize a spa discovery protocol."""
        self.transport: asyncio.DatagramTransport | None = None
        self.broadcast_handle: asyncio.TimerHandle | None = None

        self.spas: list[DiscoveredSpa] = []
        self.discovery_complete = asyncio.Event()
        self.return_once_found = return_once_found

    def broadcast(self) -> None:
        """Send a broadcast message."""
        if self.return_once_found and self.spas:  # stop broadcasting if a spa is found
            self.discovery_complete.set()
            return
        if not (transport := self.transport) or transport.is_closing():
            return  # if the transport is closed, don't broadcast

        self.transport.sendto(BROADCAST_MESSAGE, BROADCAST_ADDRESS)
        _LOGGER.debug("UDP discovery broadcast sent")

        # Re-broadcast at BROADCAST_INTERVAL
        self.broadcast_handle = asyncio.get_running_loop().call_later(
            BROADCAST_INTERVAL, self.broadcast
        )

    def connection_lost(self, exc: Exception | None) -> None:
        """Called when the connection is lost or closed."""
        if self.broadcast_handle:
            self.broadcast_handle.cancel()

    def connection_made(self, transport: asyncio.DatagramTransport) -> None:  # type: ignore[override]
        """Called when a connection is made."""
        self.transport = transport
        self.broadcast()

    def datagram_received(self, data: bytes, addr: tuple[str | Any, int]) -> None:
        """Called when some datagram is received."""
        _LOGGER.debug("Received response from %s: %s", addr[0], data)
        if b"BWGS" not in data.upper():
            return  # Unexpected response, ignore
        try:
            hostname, mac = map(str.strip, data.decode().splitlines()[:2])
            if (spa := DiscoveredSpa(*addr, mac, hostname)) not in self.spas:
                self.spas.append(spa)
            if self.return_once_found:
                self.discovery_complete.set()
        except Exception as ex:
            _LOGGER.error(ex)

    def error_received(self, exc: Exception) -> None:
        """Called when a send or receive operation raises an OSError."""
        _LOGGER.error(exc)



"""Balboa spa module."""

__version__ = "1.1.4b0"


from .client import SpaClient
from .control import EVENT_UPDATE, SpaControl
from .exceptions import SpaConnectionError

__all__ = ["SpaClient", "SpaControl", "SpaConnectionError", "EVENT_UPDATE"]



"""pybalboa exceptions."""


class SpaConnectionError(ConnectionError):
    """Spa connection could not be established."""


class SpaConfigurationNotLoadedError(Exception):
    """Raised when an operation requires a loaded spa configuration, but it has not been loaded."""

    def __init__(
        self,
        message: str = (
            "Spa configuration not loaded. "
            "Wait for async_configuration_loaded() to complete before proceeding."
        ),
    ):
        super().__init__(message)


class SpaMessageError(Exception):
    """Spa message is invalid."""



"""Balboa spa control."""

from __future__ import annotations

import logging
from collections.abc import Callable
from dataclasses import InitVar, dataclass, field
from datetime import datetime, time, timedelta
from enum import IntEnum
from typing import TYPE_CHECKING, Any, Final

from .enums import (
    ControlType,
    HeatMode,
    MessageType,
    OffLowHighState,
    OffLowMediumHighState,
    OffOnState,
    ToggleItemCode,
    UnknownState,
)

if TYPE_CHECKING:
    from .client import SpaClient

_LOGGER = logging.getLogger(__name__)


CONTROL_TYPE_MAP = {
    ControlType.PUMP: ToggleItemCode.PUMP_1,
    ControlType.BLOWER: ToggleItemCode.BLOWER,
    ControlType.MISTER: ToggleItemCode.MISTER,
    ControlType.LIGHT: ToggleItemCode.LIGHT_1,
    ControlType.AUX: ToggleItemCode.AUX_1,
    ControlType.CIRCULATION_PUMP: ToggleItemCode.CIRCULATION_PUMP,
    ControlType.TEMPERATURE_RANGE: ToggleItemCode.TEMPERATURE_RANGE,
    ControlType.HEAT_MODE: ToggleItemCode.HEAT_MODE,
}
STATE_OPTIONS_MAP: dict[int, list[IntEnum]] = {
    2: list(OffOnState),
    3: list(OffLowHighState),
    4: list(OffLowMediumHighState),
}

EVENT_UPDATE = "update"

FAULT_LOG_ERROR_CODES: Final[dict[int, str]] = {
    15: "Sensors are out of sync",
    16: "The water flow is low",
    17: "The water flow has failed",
    18: "The settings have been reset",
    19: "Priming Mode",
    20: "The clock has failed",
    21: "The settings have been reset",
    22: "Program memory failure",
    26: "Sensors are out of sync -- Call for service",
    27: "The heater is dry",
    28: "The heater may be dry",
    29: "The water is too hot",
    30: "The heater is too hot",
    31: "Sensor A Fault",
    32: "Sensor B Fault",
    34: "A pump may be stuck on",
    35: "Hot fault",
    36: "The GFCI test failed",
    37: "Standby Mode (Hold Mode)",
}


class EventMixin:
    """Event mixin."""

    _listeners: dict[str, list[Callable]] = {}

    def on(  # pylint: disable=invalid-name
        self, event_name: str, callback: Callable
    ) -> Callable:
        """Register an event callback."""
        listeners: list = self._listeners.setdefault(event_name, [])
        listeners.append(callback)

        def unsubscribe() -> None:
            """Unsubscribe listeners."""
            if callback in listeners:
                listeners.remove(callback)

        return unsubscribe

    def emit(self, event_name: str, *args: Any, **kwargs: dict[str, Any]) -> None:
        """Run all callbacks for an event."""
        for listener in self._listeners.get(event_name, []):
            listener(*args, **kwargs)


class SpaControl(EventMixin):
    """Spa control."""

    _options: list[IntEnum]

    def __init__(
        self,
        client: SpaClient,
        control_type: ControlType,
        states: int | list[IntEnum] = 1,
        index: int | None = None,
        custom_options: list[IntEnum] | None = None,
    ) -> None:
        """Initialize a spa control."""
        self._client = client
        self._control_type = control_type
        self._index = index

        self._name = f"{control_type.value}{'' if index is None else f' {index + 1}'}"
        self._code = CONTROL_TYPE_MAP[control_type]
        self._state_value = UnknownState.UNKNOWN.value
        self._state: IntEnum = UnknownState.UNKNOWN

        if isinstance(states, int):
            self._states = states
            self._options = STATE_OPTIONS_MAP[states]
        else:
            self._states = len(states)
            self._options = states
        self._custom_options = custom_options

    def __repr__(self) -> str:
        """Return repr(self)."""
        return f"{self.name}: {self.state.name}"

    @property
    def client(self) -> SpaClient:
        """Return the client."""
        return self._client

    @property
    def control_type(self) -> ControlType:
        """Return the control type."""
        return self._control_type

    @property
    def index(self) -> int | None:
        """Return the index."""
        return self._index

    @property
    def name(self) -> str:
        """Return the control name."""
        return self._name

    @property
    def options(self) -> list[IntEnum]:
        """Return the available control options."""
        return self._custom_options or [*self._options]

    @property
    def state(self) -> IntEnum:
        """Get the control's current state."""
        return self._state

    def update(self, state: int) -> None:
        """Update the control's current state."""
        if self._state_value != state:
            self._state_value = state
            self._state = next(
                (option for option in self._options if option == state),
                self._options[-1]
                if self.control_type == ControlType.PUMP and state >= self._states
                else UnknownState.UNKNOWN,
            )
            _LOGGER.debug(
                "%s -- %s is now %s (%s)",
                self._client.host,
                self.name,
                self.state.name,
                state,
            )
            self.emit(EVENT_UPDATE)

    async def set_state(self, state: int | IntEnum) -> bool:
        """Set control to state."""
        if state not in self.options:
            _LOGGER.error("Cannot set state to %s", state)
            return False
        if self._state == state:
            return True
        min_toggle = 1
        if self._state != UnknownState.UNKNOWN:
            min_toggle = max((state - self._state) % self._states, 1)
        for _ in range(min_toggle):
            await self._client.send_message(
                MessageType.TOGGLE_STATE, self._code + (self._index or 0)
            )
        return True


class HeatModeSpaControl(SpaControl):
    """Heat mode spa control."""

    def __init__(self, client: SpaClient) -> None:
        """Initialize a heat mode spa control."""
        super().__init__(
            client,
            ControlType.HEAT_MODE,
            list(HeatMode),
            custom_options=[*HeatMode][:2],
        )

    async def set_state(self, state: int | HeatMode) -> bool:
        """Set control to state."""
        if state not in self.options:
            _LOGGER.error("Cannot set state to %s", state)
            return False
        if self._state == state:
            return True
        i = 2 if self.state == HeatMode.READY_IN_REST and state == HeatMode.READY else 1
        for _ in range(i):
            await self._client.send_message(MessageType.TOGGLE_STATE, self._code)
        return True


@dataclass
class FaultLog:
    """Fault log."""

    count: int
    entry_number: int
    message_code: int
    days_ago: int
    time_hour: int
    time_minute: int
    flags: int
    target_temperature: int
    sensor_a_temperature: int
    sensor_b_temperature: int

    current_time: InitVar[datetime | None] = None
    fault_datetime: datetime = field(init=False)

    def __post_init__(self, current_time: datetime | None) -> None:
        """Compute the fault datetime on initialization."""
        self.fault_datetime = datetime.combine(
            (current_time or datetime.now()) - timedelta(days=self.days_ago),
            time(self.time_hour, self.time_minute),
        )

    def __str__(self) -> str:
        """Return str(self)."""
        return (
            f"Fault log {self.entry_number + 1}/{self.count}: {self.message} "
            f"occurred on {self.fault_datetime:%Y-%m-%d at %H:%M}"
        )

    @property
    def message(self) -> str:
        """Return the message for the error code."""
        return FAULT_LOG_ERROR_CODES.get(
            self.message_code, f"Unknown ({self.message_code})"
        )
